"""Operators
============

Information regarding mesh, mathematical operators, and domain decomposition.

"""
import inspect
import itertools
import math
import sys
from collections import OrderedDict
from math import pi

from .log import logger
from .util import docstring_params


def _str_len(length):
    """Generates a special string if length is multiple of pi. If not provide 3
    decimal places past floating point.

    """
    if (length / pi).is_integer():
        str_len = repr(int(length / pi)) + "pi"
    else:
        str_len = f"{length:.3f}".rstrip("0")

    return str_len


def next_power(value, base=2):
    """Compute the perfect power of ``base`` greater than or equal to ``value``.

    :param value:
    :type value: int or float

    :param int base:

    :return int:

    """
    exponent = math.ceil(math.log(value, base))
    return base**exponent


class Operators:
    """Container for parameters and writing :ref:`box <nek:mesh_gen>` and
    :ref:`SIZE <nek:case_files_size>` files.

    .. note:: Some values are not available as parameters and instead
              automatically computed for generating the SIZE file.

    ==============  ======================== ==================================
    SIZE            `properties`             Comment
    ==============  ======================== ==================================
    ``lelg``        :any:`max_n_seq`         Max. number of elements globally
    ``lelt``        :any:`max_n_loc`         | Max. number of elements per
                                               processor (should be not smaller
                                             | than ``lelg/lpmin``, i.e.

    ``lelx``        :any:`max_nx`            | **Automatically computed**. Max.
                                               number of elements along x
                                             | direction for the global tensor
                                               product solver / dimensions.

    ``lely``        :any:`max_ny`            Same as above for ``y`` direction.
    ``lelz``        :any:`max_nz`            Same as above for ``z`` direction.

    ``lorder``      :any:`max_order_time`    | Max. temporal order
                                             | **Automatically computed** based
                                             | ``params.nek.general.\
time_stepper``

    ``lbelt``       :any:`order_mhd`         | **Automatically computed** as
                                             |  ``lelt`` if ``"MHD" in
                                               params.nek.problem_type.equation``.
    ``lpelt``       :any:`order_linear`      | **Automatically computed** as
                                             |  ``lelt`` if ``"linear" in
                                               params.nek.problem_type.equation``.
    ``lcvelt``      :any:`order_cvode`       | **Automatically computed** as
                                             |  ``lelt`` if
                                               ``params.nek.cvode._enabled is
                                               True``

    ``lx1m``        :any:`order_mesh_solver` | p-order for mesh solver.
                                               **Automatically computed** based
                                             | ``params.nek.general.stress\
_formulation`` and whether
                                             | Arbitrary Lagrangian-Eulerian
                                               (ALE) methods are used or not.
    ==============  ======================== ==================================

    """

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params.oper* container."""
        attribs = {
            "nx": 8,
            "ny": 8,
            "nz": 8,
            "origin_x": 0.0,
            "origin_y": 0.0,
            "origin_z": 0.0,
            "ratio_x": 1.0,
            "ratio_y": 1.0,
            "ratio_z": 1.0,
            "Lx": 2 * pi,
            "Ly": 2 * pi,
            "Lz": 2 * pi,
            "boundary": ["P", "P", "W", "W", "P", "P"],
            "boundary_scalars": [],
            "dim": 3,
            "nproc_min": 4,
            "nproc_max": 32,
            "scalars": 1,
        }
        params._set_child("oper", attribs=attribs)
        params.oper._set_doc(
            """
Parameters for mesh description:

- ``nx``, ``ny``, ``nz``: int
    Number of elements in each directions
- ``origin_x``, ``origin_y``, ``origin_z``: float
    Starting coordinate of the mesh (default: 0.0)
- ``ratio_x``, ``ratio_y``, ``ratio_z``: float
    Mesh stretching ratio (default: 1.0)
- ``Lx``, ``Ly``, ``Lz``: float
    Length of the domain

Parameters for boundary conditions:

- ``boundary``: list[str]
    :ref:`Velocity boundary conditions <nek:tab:ubcf>`
- ``boundary_scalars``: list[str]
    :ref:`Temperature and passive scalar boundary conditions <nek:tab:userbct>`

The following table matches counterpart of mandatory :ref:`SIZE
<nek:case_files_size>` variables.

==========  ===================   =============================================
SIZE        params.oper           Comment
==========  ===================   =============================================
``ldim``    ``dim``               Domain dimensions (2 or 3)

``lpmin``   ``nproc_min``         Min MPI ranks
``lpmax``   ``nproc_max``         Max MPI ranks
``ldimt``   ``scalars``           Number of auxilary fields (minimum 1).

==========  ===================   =============================================

"""
        )

        attribs = {
            option: 1
            for option in (
                "hist",
                "obj",
                "perturb",
                "scalars_cons",
                "scalars_proj",
                "sessions",
            )
        }
        attribs["dim_proj"] = 20
        attribs["dim_krylov"] = 30

        params.oper._set_child("max", attribs=attribs)
        params.oper.max._set_doc(
            """
The following table matches counterpart of optional :ref:`SIZE
<nek:case_files_size>` variables. These refer to upper bound number of
`something`. The parameters are considered "optional" and would be ignored with
the default values.

==============  ===================   =========================================
SIZE            params.oper.max       Comment
==============  ===================   =========================================
``mxprev``      ``dim_proj``          Max. dimension of projection space
``lgmres``      ``dim_krylov``        Max. dimension of Krylov space for GMRES
``lhis``        ``hist``              Max. number of history (i.e. monitoring)
                                      points.

``maxobj``      ``obj``               Max. number of objects?

``lpert``       ``perturb``           Max. number of perturbations
``toteq``       ``scalars_cons``      Max. conserved scalars
``ldimt_proj``  ``scalars_proj``      Max. scalars for residual projection
``nsessmax``    ``sessions``          Max. sessions to NEKNEK

==============  ===================   =========================================

"""
        )

        attribs = {
            "order": 6,
            "order_out": 6,
            "coef_dealiasing": 2.0 / 3,
            "staggered": "auto",
        }
        params.oper._set_child("elem", attribs=attribs)
        params.oper.elem._set_doc(
            r"""
The following table relate to element configuration for certain operations.
The parameters are considered "optional" (except for ``lx1`` / ``order`` and
``lxd`` / ``coef_dealiasing`` which are mandatory) and would be ignored with
the default values.

==========      ===================   =========================================
SIZE            params.oper.elem      Comment
==========      ===================   =========================================
``lxd``         ``coef_dealiasing``   | p-order [#f1]_ for over-integration.
                                        **Automatically computed** from float
                                      | ``coef_dealiasing``. See
                                        :any:`order_dealiasing`

``lx1``         ``order``             p-order (avoid uneven and values <6).

``lxo``         ``order_out``         Max. p-order on output (should be ``>=
                                      order``. See :any:`order_out`)

``lx2``         ``staggered``         | p-order for pressure. **Automatically
                                        computed**. See :any:`order_pressure`

==========      ===================   =========================================

.. [#f1] Polynomial order which means the number of GLL / GL points per element.

"""
        )
        attribs = {"fast_diag": False}
        params.oper._set_child("misc", attribs=attribs)
        params.oper.misc._set_doc(
            r"""
Other miscellaneous parameters:

==========      ===================   =========================================
SIZE            params.oper.misc      Comment
==========      ===================   =========================================
``lfdm``        ``fast_diag``         | Equals to True for global tensor
                                        product solver (that uses fast
                                      | diagonalization method). ``False``
                                        otherwise.
==========      ===================   =========================================

"""
        )

    def __init__(self, sim=None, params=None):
        self.params = sim.params if sim else params
        self.axes = ("x", "y", "z")

    @property
    def max_n_seq(self):
        """Equivalent to ``lelg``."""
        params = self.params
        if params.oper.dim == 2:
            return next_power(params.oper.nx * params.oper.ny)
        else:
            return next_power(params.oper.nx * params.oper.ny * params.oper.nz)

    @property
    def max_n_loc(self):
        """Equivalent to ``lelt``. The next integer greater than or equals
        ``max_n_seq / params.oper.nproc_min``.
        """
        return math.ceil(self.max_n_seq / self.params.oper.nproc_min)

    @property
    def max_nx(self):
        """Equivalent to ``lelx``."""
        return next_power(self.params.oper.nx)

    @property
    def max_ny(self):
        """Equivalent to ``lely``."""
        return next_power(self.params.oper.ny)

    @property
    def max_nz(self):
        """Equivalent to ``lelz``."""
        return next_power(self.params.oper.nz)

    @property
    def max_order_time(self):
        """Equivalent to ``lorder``."""
        (
            _,
            time_scheme,
            order_time,
        ) = self.params.nek.general.time_stepper.upper().partition("BDF")
        if time_scheme != "BDF":
            raise ValueError("Unsupported time integration scheme")

        return int(order_time)

    @property
    def nb_fields(self):
        """Used in .box file."""
        return self.params.oper.scalars + 1

    @property
    def order(self):
        r"""Equivalent to ``lx1``. Controls the polynomial order of the
        velocity field.

        .. note::

            True polynomial order of the element is given by
            :math:`\mathbb{P}_N` = ``lx1`` - 1 = :any:`order` - 1

        """
        return self.params.oper.elem.order

    @property
    def order_out(self):
        """Equivalent to ``lxo``."""
        elem = self.params.oper.elem
        if elem.order_out < elem.order:
            raise ValueError(
                f"Max GLL points on output should not be less than element "
                f"order. {elem.order_out} < {elem.order}"
            )

        return elem.order_out

    @property
    def order_dealiasing(self):
        """Equivalent to ``lxd``."""
        elem = self.params.oper.elem
        return math.ceil(elem.order / elem.coef_dealiasing)

    @property
    def order_pressure(self):
        r"""Equivalent to ``lx2``. Controls the order for the pressure field.

        .. note::

            The property :any:`order_pressure` is determined by
            the value of :attr:`params.oper.elem.staggered`.

            - If staggered == "auto":

              * If "lin" in problemtype_equation
                :math:`\implies \mathbb{P}_{N-2}`
              * Else
                :math:`\implies \mathbb{P}_N`

            - If staggered is True :math:`\implies \mathbb{P}_{N-2}`
            - If staggered is False :math:`\implies \mathbb{P}_N`

        """

        pn = self.order
        staggered = self.params.oper.elem.staggered

        problemtype_equation = self.params.nek.problemtype.equation.lower()

        if "lin" in problemtype_equation and staggered is False:
            logger.warning(
                "Linear equation type and staggered == False leads to "
                "undefined behaviour in Nek5000.  User should put "
                'params.oper.elem.staggered = True or "auto" to have evolution'
                "of perturbation field."
            )

        if staggered == "auto":
            if "lin" in problemtype_equation:
                return pn - 2
            else:
                return pn
        elif staggered is True:
            return pn - 2
        elif staggered is False:
            return pn
        else:
            raise ValueError(
                'params.nek have to be in [True, False, "auto"]. '
                f"staggered = {staggered}"
            )

    @property
    def order_mesh_solver(self):
        """
        Equivalent to ``lx1m``.

        .. todo:: Must include a condition to check if ALE methods are used or
                  not.

        """
        return (
            self.order
            if (
                self.params.nek.problemtype.stress_formulation
                # or ALE
            )
            else 1
        )

    @property
    def order_mhd(self):
        """Equivalent to ``lbelt``."""
        return (
            self.max_n_loc
            if "mhd" in self.params.nek.problemtype.equation.lower()
            else 1
        )

    @property
    def order_linear(self):
        """Equivalent to ``lpelt``."""
        return (
            self.max_n_loc
            if "lin" in self.params.nek.problemtype.equation.lower()
            else 1
        )

    @property
    def order_cvode(self):
        """Equivalent to ``lcvelt``."""
        return self.max_n_loc if self.params.nek.cvode._enabled else 1

    def memory_required(self):
        """According to Nek5000 :ref:`nek:faq` the following estimate is made

        ::

              lx1*ly1*lz1*lelt * 3000byte + lelg * 12byte + MPI + optional libraries
              (e.g. CVODE)

        Returns
        -------
        memory_required: int
            in bytes

        """
        params = self.params
        elem = params.oper.elem

        lx1 = elem.order
        ldim = params.oper.dim
        ly1 = lx1
        lz1 = 1 + (ldim - 2) * (lx1 - 1)
        lelt = self.max_n_loc
        lelg = self.max_n_seq

        return lx1 * ly1 * lz1 * lelt * 3000 + lelg * 12

    def _str_Ln(self):
        params = self.params.oper
        dim = params.dim
        str_l = map(_str_len, (params.Lx, params.Ly, params.Lz)[:dim])
        str_n = map(str, (params.nx, params.ny, params.nz)[:dim])
        return str_l, str_n

    def _modify_sim_repr_maker(self, sim_repr_maker):
        repr_oper = self.produce_str_describing_oper()
        sim_repr_maker.add_word(repr_oper)

    def produce_str_describing_oper(self):
        """Produce a string describing the mesh volume and number of
        elements.

        """
        str_L, str_n = self._str_Ln()
        return f"{'x'.join(str_n)}_V{'x'.join(str_L)}"

    def produce_long_str_describing_oper(self, oper_method="Base"):
        """Produce a multi-line string describing the mesh volume and number of
        elements.

        """
        str_L, str_n = self._str_Ln()
        string = ""
        for key, value in zip(("Lx", "Ly", "Lz"), str_L):
            string += f"- {key} = {value}\n"
        for key, value in zip(("nx", "ny", "nz"), str_n):
            string += f"- {key} = {value}\n"
        return f"Nek5000 operator:\n{string}"

    def info_box(self, comments=""):
        """Gather information for writing a box file.

        Returns
        -------
        dict

        """
        params = self.params
        comments += """
Autogenerated using snek5000.operators.Operators.write_box()

If dim < 0 .re2 file will be generated

If nelx (y or z) < 0, then genbox automatically generates the
                      grid spacing in the x (y or z) direction
                      with a geometric ratio given by "ratio".
                      ( ratio=1 implies uniform spacing )

Note that the values for "x0 x1 ratio" _must_ be formatted as `.4f`.

Note that each code for the boundary cond. _must_ have 3 spaces.
"""

        def _str_grid(*args):
            fmt = "{:.4f} {:.4f} {:.4f}"
            args = (float(value) for value in args)
            return fmt.format(*args)

        dim = params.oper.dim
        boundary = params.oper.boundary
        boundary_scalars = params.oper.boundary_scalars

        for bc in itertools.chain(boundary, boundary_scalars):
            if len(bc) > 3:
                raise ValueError(
                    f"Length of boundary condition {bc} shall not exceed 3 characters"
                )

        # A dictionary mapping a comment to grid
        grid_info = OrderedDict(
            [
                (
                    "nelx nely nelz",
                    " ".join(
                        str(-n)
                        for n in (params.oper.nx, params.oper.ny, params.oper.nz)[:dim]
                    ),
                ),
                (
                    "x0 x1 ratio",
                    _str_grid(
                        params.oper.origin_x, params.oper.Lx, params.oper.ratio_x
                    ),
                ),
                (
                    "y0 y1 ratio",
                    _str_grid(
                        params.oper.origin_y, params.oper.Ly, params.oper.ratio_y
                    ),
                ),
            ]
        )

        if params.oper.dim == 3:
            grid_info.update(
                [
                    (
                        "z0 z1 ratio",
                        _str_grid(
                            params.oper.origin_z,
                            params.oper.Lz,
                            params.oper.ratio_z,
                        ),
                    ),
                ]
            )

        if boundary:
            grid_info.update(
                [
                    (
                        "Velocity BCs",
                        ",".join(bc.ljust(3) for bc in boundary),
                    )
                ]
            )

        if boundary_scalars:
            grid_info.update(
                [
                    (
                        "Temperature / scalar BCs",
                        ",".join(bc.ljust(3) for bc in boundary_scalars),
                    )
                ]
            )

        info = {
            "comments": comments,
            "dim": str(-params.oper.dim),
            "grid_info": grid_info,
            "nb_fields": str(self.nb_fields),  # scalars + velocity
        }
        return info

    def write_box(self, template, fp=sys.stdout, comments="", **template_vars):
        """Write the .box file which is input for the ``genbox`` meshing
        tool.

        Parameters
        ----------
        template : jinja2.environment.Template
            Template instance loaded from something like ``box.j2``
        fp : io.TextIOWrapper
            File handler to write to
        comments: str
            Comments on top of the box file
        template_vars: dict
            Keyword arguments passed while rendering the Jinja templates

        """
        template_vars.update(self.info_box(comments))

        # Write the box file
        output = template.render(**template_vars)
        fp.write(output)

    def write_size(self, template, fp=sys.stdout, comments="", **template_vars):
        """Write the SIZE file which is required for compilation.

        Parameters
        ----------
        template : jinja2.environment.Template
            Template instance loaded from something like ``SIZE.j2``
        fp : io.TextIOWrapper
            File handler to write to
        comments: str
            Comments on top of the SIZE file
        template_vars: dict
            Keyword arguments passed while rendering the Jinja templates

        """
        comments += """
Autogenerated using snek5000.operators.Operators.write_size()
"""
        template_vars.update({"comments": comments, "params": self.params})
        # Include all @property attributes to the template
        template_vars.update(
            {
                name: getattr(self, name)
                for name, _ in inspect.getmembers(
                    self.__class__, lambda attr: isinstance(attr, property)
                )
            }
        )
        output = template.render(**template_vars)
        fp.write(output)


Operators.__doc__ += """\n

.. note:: We deliberately try not to use the variable names used in Nek5000, as
          those are ambiguously named.


""" + docstring_params(
    Operators, indent_len=0
)



import sys

import pytest

from snek5000 import load


def test_load(sim_canonical):
    # Reload snek5000_canonical so that the load method would freshly import snek5000_canonical
    # https://stackoverflow.com/a/57851153
    for module in tuple(sys.modules):
        if module.startswith("snek5000_canonical"):
            del sys.modules[module]

    sim2 = load(sim_canonical.path_run, reader="pymech_avg")
    with pytest.raises(IOError, match="no files to open"):
        sim2.output.phys_fields.plot_mean_vel()


@pytest.mark.slow
def test_package_canonical(sim_canonical):
    sim = sim_canonical
    assert sim.make.exec("compile")



import pytest
import xarray as xr

from snek5000 import load
from snek5000.util import repeat


@pytest.mark.slow
def test_history_points(sim_cbox_executed):
    sim = sim_cbox_executed
    params = sim.params
    p_hist = params.output.history_points

    hist_points = sim.output.history_points
    coords, df = hist_points.load()

    assert len(coords) == len(p_hist.coords)
    assert len(df) == 1 + params.nek.general.num_steps // p_hist.write_interval

    hist_points.plot("temperature")

    # let's tweak the cbox.his to mimic a new data point
    with open(hist_points.path_file, "r") as file:
        file.readline()
        for _ in repeat(len(p_hist.coords)):
            file.readline()

        line_data = file.readline()
    with open(hist_points.path_file, "a") as file:
        file.write(f"{line_data}\nbad\n")

    hist_points.plot_1point(0, "temperature", tmin=0, tmax=1)


@pytest.mark.slow
def test_loadsimul_phys_fields(sim_cbox_executed):
    sim = load(sim_cbox_executed.path_run)

    ux = sim.output.phys_fields.get_var("ux")
    assert isinstance(ux, xr.DataArray)


@pytest.mark.slow
def test_phys_fields_get_var_before_load(sim_cbox_executed):
    sim = sim_cbox_executed
    sim.output.phys_fields.init_reader()
    ux = sim.output.phys_fields.get_var("ux")
    assert isinstance(ux, xr.DataArray)


@pytest.mark.slow
def test_print_stdout(sim_cbox_executed):
    sim = sim_cbox_executed

    sim.output.print_stdout.plot_nb_iterations()
    sim.output.print_stdout.plot_dt_cfl()



"""resources for building
======================

Contains common snakemake rules and jinja templates.

Snakemake rules
---------------

- ``nek5000.smk``: subworkflow_ for building Nek5000 tools

.. literalinclude:: ../../src/snek5000/resources/nek5000.smk
    :language: python

- ``compiler.smk``: module_ for compiling user code

.. literalinclude:: ../../src/snek5000/resources/compiler.smk
    :language: python

- ``io.smk``: module_ for cleaning and archiving simulation artefacts

.. literalinclude:: ../../src/snek5000/resources/io.smk
    :language: python

- ``internal.smk``: module_ for generating simulation files, serves as inputs
  to Snakemake rules above

.. literalinclude:: ../../src/snek5000/resources/internal.smk
    :language: python

.. _module: https://snakemake.readthedocs.io/en/stable/snakefiles/modularization.html#modules
.. _subworkflow: https://snakemake.readthedocs.io/en/stable/snakefiles/modularization.html#sub-workflows


Jinja templates
---------------

- ``box.j2``: See also
    :any:`snek5000.operators.Operators.write_box`,
    :any:`snek5000.output.base.Output.write_box`.
    :ref:`nek:mesh_gen`

.. literalinclude:: ../../src/snek5000/resources/box.j2
    :language: jinja

- ``makefile_usr.inc.j2``: See also
    :any:`snek5000.output.base.Output.write_makefile_usr`.

.. literalinclude:: ../../src/snek5000/resources/makefile_usr.inc.j2
    :language: jinja

- ``SIZE.j2``: See also
    :any:`snek5000.operators.Operators.write_size`,
    :any:`snek5000.output.base.Output.write_size`,
    Nek5000 docs on :ref:`nek:case_files_size` file.

.. literalinclude:: ../../src/snek5000/resources/SIZE.j2
    :language: jinja

- ``compile.sh.j2``: See also
    :any:`snek5000.output.base.Output.write_compile_sh`.

.. literalinclude:: ../../src/snek5000/resources/compile.sh.j2
    :language: jinja

"""


import jinja2


def get_base_templates():
    """Get templates (box, makefile_usr and size) from ``snek5000.resources``."""
    env = jinja2.Environment(
        loader=jinja2.PackageLoader("snek5000", "resources"),
        undefined=jinja2.StrictUndefined,
    )

    box = env.get_template("box.j2")
    makefile_usr = env.get_template("makefile_usr.inc.j2")
    size = env.get_template("SIZE.j2")

    return box, makefile_usr, size



from tuto_cbox import Simul, params

params.output.sub_directory = "examples_snek/movie"

aspect_ratio = 0.7
params.prandtl = 0.71
params.Ra_side = 4e9
nb_elements = 28
order = 10

Ly = params.oper.Ly
params.oper.Lx = Ly / aspect_ratio

params.oper.ny = ny = nb_elements
nx = params.oper.nx = int(nb_elements / aspect_ratio)

params.oper.elem.order = params.oper.elem.order_out = order

params.short_name_type_run = f"Ra{params.Ra_side:.3e}_{nx*order}x{ny*order}"

params.nek.general.write_interval = 0.5
params.nek.general.end_time = 100
params.nek.general.target_cfl = 1.7

params.nek.velocity.residual_tol = 1e-08
params.nek.pressure.residual_tol = 1e-06

sim = Simul(params)

sim.make.exec("run_fg", nproc=4)

print(
    f"""You can run:

cd {sim.output.path_run}; snek-ipy-load

Then, from IPython:

from functools import partial

animate = partial(
    sim.output.phys_fields.animate, dt_frame_in_sec=0.15, dt_equations=0.5,
    tmax=100,
    pcolor_kw=dict(cmap="coolwarm"),
    quiver_kw=dict(scale=12),
    normalize_vectors=False,
    fig_kw=dict(figsize=(6, 4), dpi=70),
)

animate(interactive=True)
animate(save_file="movie.mp4")

"""
)



import snek5000


def test_get_nek_source_root():
    assert snek5000.get_nek_source_root()


def test_resource():
    assert snek5000.get_snek_resource("nek5000.smk")
    assert snek5000.get_snek_resource("default_configfile.yml")



"""Utilities to restart a simulation
====================================

"""
import os
import sys
from enum import Enum
from pathlib import Path
from textwrap import dedent

from fluidsim_core.scripts.restart import RestarterABC

from ..log import logger
from ..output import _make_path_session, _parse_path_run_session_id
from ..params import load_params
from ..solvers import get_solver_short_name, import_cls_simul
from .files import _path_try_from_fluidsim_path, next_path


class SnekRestartError(Exception):
    pass


class SimStatus(Enum):
    """Simulation status inspired from HTTP response status codes_.

    .. _codes: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status

    """

    OK = (200, "OK: All prerequisities satisfied to restart.")
    RESET_CONTENT = (
        205,
        (
            "Reset Content: Multi-file restart found. Some field files exist. "
            "Restarting in the same session would overwrite files. "
            "Ensure current session is archived or restart in a new session or "
            "a new directory."
        ),
    )
    PARTIAL_CONTENT = (
        206,
        (
            "Partial Content: No multi-file restart found. Some field files exist. "
            "Ensure current session is archived or restart in a new session or "
            "a new directory."
        ),
    )
    NOT_FOUND = (
        404,
        "Not Found: SIZE and/or nek5000 is missing.",
    )
    LOCKED = (
        423,
        (
            "Locked: The path is currently locked by snakemake. "
            "Execute `snakemake --unlock` function snek5000.make.unlock."
        ),
    )
    TOO_EARLY = (
        425,
        "Too Early: Seems like snakemake was never executed.",
    )

    def __init__(self, code: int, message: str):
        self.code = code  #: status code
        self.message = message  #: helpful description


def get_status(path_dir, session_id=None, verbose=False):
    """Get status of a simulation run by verifying its contents. It checks if:

    - snakemake was ever executed
    - directory is locked by snakemake (either due to a running
      simulation or a terminated one)
    - necessary files for starting a simulation exist
    - restart files exist

    Parameters
    ----------
    path : str or path-like
        Path to an existing simulation directory
    session_id : int
        Integer suffix of the session directory
    verbose : bool
        Print out the path and its contents

    Returns
    -------
    _ : SimStatus
        Enumeration indicating status code and message

    """
    path = Path(path_dir)
    if session_id:
        path_session = _make_path_session(path, session_id)
    else:
        path_session = Path(load_params(path_dir).output.path_session)

    locks_dir = path / ".snakemake" / "locks"
    contents = os.listdir(path)

    if verbose:
        print(path, "\nContents:", contents)

    if not (path / ".snakemake").exists():
        return SimStatus.TOO_EARLY
    elif locks_dir.exists():
        locks = tuple(locks_dir.iterdir())
        if locks:
            return SimStatus.LOCKED

    if not {"SIZE", "nek5000"}.issubset(contents):
        return SimStatus.NOT_FOUND

    checkpoints = set(path.glob("rs6*0.f?????"))
    field_files = set(path_session.glob("*0.f?????"))

    if checkpoints and field_files:
        return SimStatus.RESET_CONTENT
    elif field_files:
        return SimStatus.PARTIAL_CONTENT
    else:
        return SimStatus.OK


def load_for_restart(
    path_dir=".",
    use_start_from=None,
    use_checkpoint=None,
    session_id=None,
    verify_contents=True,
    new_dir_results=False,
    only_check=False,
):
    """Load params and Simul for a restart.

    >>> params, Simul = load_for_restart(name_dir)

    Parameters
    ----------

    path_dir: str or path-like (optional)
        The directory of an existing simulation. If nothing is given, we load the
        data in the current directory.  Can be an absolute path, a relative path,
        or even simply just the name of the directory under $FLUIDSIM_PATH.

    use_start_from: str or int
        Name or index of the field file to restart from. Mutually exclusive option
        with ``use_checkpoint``.

    use_checkpoint: int, {1, 2}
        Number of the multi-file checkpoint file set to restart from. Mutually
        exclusive parameter with ``use_start_from``.

    session_id: int
        Indicate which session directory should be used to look for restart files.
        If not specified it would default to the `path_session` value last
        recorded in the `params_simul.xml` file.

    verify_contents: bool
        Verify directory contents to avoid runtime errors.

    new_dir_results: bool (default False)
        Create a new directory for the new simulation.

    Notes
    -----
    How it works:

    - If ``verify contents`` is `True`, do so using :func:`get_status`
    - Reads ``params_simul.xml`` if it exists, and if not falls back to ``.par`` file.
    - Modifies parameters (in memory, but does not write into the filesystem,
      yet) ``start_from`` (Nek5000) or checkpoint module (requires KTH
      framework) with appropriate ``chkp_fnumber`` to restart from.

    """
    path = _path_try_from_fluidsim_path(path_dir)

    # In case the user specifies the path to a session sub-directory
    if session_id is None:
        path, session_id = _parse_path_run_session_id(path)

    try:
        params = load_params(path)
    except (ValueError, OSError) as err:
        raise SnekRestartError(err) from err

    status = get_status(path, session_id or params.output.session_id)

    if verify_contents:
        if status.code >= 400:
            raise SnekRestartError(f"{status.code}: {status.message}")
        else:
            logger.info(f"{status.code}: {status.message}")

    # Load Simul class and parameters
    short_name = get_solver_short_name(path)

    try:
        Simul = import_cls_simul(short_name)
    except ImportError:
        raise ImportError(f"Cannot import Simul class of solver {short_name}")

    # Set restart file
    if use_start_from and use_checkpoint:
        raise SnekRestartError(
            "Options use_start_from and use_checkpoint are mutually exclusive. "
            "Use only one option at a time."
        )
    elif not use_start_from and not use_checkpoint:
        raise SnekRestartError(
            "No restart files were requested. "
            "This would result in a fresh simulation in a new session."
        )
    elif use_checkpoint:
        if use_checkpoint in (1, 2) and status in (
            SimStatus.OK,
            SimStatus.RESET_CONTENT,
        ):
            params.nek.chkpoint.chkp_fnumber = use_checkpoint
            params.nek.chkpoint.read_chkpt = True
        else:
            raise SnekRestartError(
                f"Restart checkpoint {use_checkpoint} is invalid / does not exist"
            )

    if hasattr(params, "output") and hasattr(params.output, "HAS_TO_SAVE"):
        params.output.HAS_TO_SAVE = True

    params.NEW_DIR_RESULTS = bool(new_dir_results)

    if use_start_from:
        if session_id is not None:
            old_path_session = _make_path_session(path, session_id)
        else:
            old_path_session = Path(params.output.path_session)
        try:
            index_start_from = int(use_start_from)
        except ValueError:
            path_start_from = old_path_session / use_start_from
        else:
            paths = sorted(old_path_session.glob(f"{short_name}0.*"))
            path_start_from = paths[index_start_from]
        params.nek.general._set_internal_attr("_path_start_from", path_start_from)

    name_restart_file = "init_state.restart"
    if new_dir_results:
        params.path_run = None
        params.output.path_session = None
        params.output.session_id = 0
        if use_start_from:
            params.nek.general.start_from = name_restart_file
            # new option Nek5000 master for interpolation on a new mesh
            # params.nek.general.start_from = name_restart_file + " int"
    else:
        new_session_id, new_path_session = next_path(
            path / "session", force_suffix=True, return_suffix=True
        )
        params.output.session_id = new_session_id
        params.output.path_session = new_path_session
        if not only_check:
            new_path_session.mkdir(exist_ok=True)

        if not only_check and use_start_from:
            if path_start_from.exists():
                params.nek.general.start_from = name_restart_file
                src = f"../{old_path_session.name}/{path_start_from.name}"
                dest = new_path_session / name_restart_file
                logger.debug(f"Symlinking {dest} -> {src}")
                dest.symlink_to(src)
            else:
                raise SnekRestartError(f"Restart file {path_start_from} not found")

    return params, Simul


class Restarter(RestarterABC):
    def create_parser(self):
        parser = super().create_parser()

        parser.add_argument(
            "-np",
            "--nb-mpi-procs",
            type=int,
            default=4,
            help="Number of MPI processes",
        )
        parser.add_argument(
            "--use-start-from",
            type=str,
            default=None,
            help=(
                "Name (relative to the session path) of the field file "
                "to restart from. "
                "Mutually exclusive option with `use_checkpoint`."
            ),
        )
        parser.add_argument(
            "--use-checkpoint",
            type=int,
            default=None,
            help=(
                "Number of the multi-file checkpoint file set to restart from. "
                "Mutually exclusive parameter with `use_start_from`."
            ),
        )
        parser.add_argument(
            "--session-id",
            type=int,
            default=None,
            help=(
                "Indicate which session directory should be used to look for "
                "restart files. If not specified it would default to the "
                "`path_session` value last recorded in the `params_simul.xml` file."
            ),
        )
        parser.add_argument(
            "--skip-verify-contents",
            action="store_true",
            help="Do not verify directory contents to avoid runtime errors.",
        )
        parser.add_argument(
            "--add-to-end-time",
            type=float,
            default=None,
            help="Time added to params.nek.general.end_time",
        )
        parser.add_argument(
            "--end-time",
            type=float,
            default=None,
            help="params.nek.general.end_time",
        )
        parser.add_argument(
            "--num-steps",
            type=int,
            default=None,
            help="params.nek.general.num_steps",
        )

        return parser

    _str_command_after_simul = dedent(
        """
        # To visualize with IPython:

        cd {path_run}; snek-ipy-load
    """
    )

    def _get_params_simul_class(self, args):
        if args.use_start_from is None and args.use_checkpoint is None:
            logger.error("Either --use-start-from or --use-checkpoint have to be given")
            sys.exit(1)
        return load_for_restart(
            args.path,
            use_start_from=args.use_start_from,
            use_checkpoint=args.use_checkpoint,
            session_id=args.session_id,
            verify_contents=not args.skip_verify_contents,
            new_dir_results=args.new_dir_results,
            only_check=args.only_check,
        )

    def _set_params_time_stepping(self, params, args):
        if args.num_steps is not None:
            params.nek.general.stop_at = "numSteps"
            params.nek.general.num_steps = int(args.num_steps)
        elif args.end_time is not None or args.add_to_end_time is not None:
            params.nek.general.stop_at = "endTime"
            if args.end_time is not None:
                end_time = args.end_time
            else:
                end_time = float(params.nek.general.end_time) + args.add_to_end_time
            params.nek.general.end_time = end_time

    def _start_sim(self, sim, args):
        if args.new_dir_results:
            if args.use_start_from:
                sim.create_symlink_start_from_file(
                    sim.params.nek.general._path_start_from
                )
            elif args.use_checkpoint:
                sim.create_symlinks_checkpoint_files(args.path)
        sim.make.exec("run_fg", nproc=args.nb_mpi_procs)

    def _check_params_time_stepping(self, params, path_file, args):
        args_times = [args.num_steps, args.end_time, args.add_to_end_time]
        if sum(arg is not None for arg in args_times) > 1:
            raise ValueError(
                "--add-to-end-time, --end-time and --num-steps are exclusive options."
            )

    def _get_path_restart_file(self, params, args):
        if args.use_start_from is not None:
            path_file = args.use_start_from
        elif args.use_checkpoint is not None:
            path_file = f"Use checkpoint files (use_checkpoint={args.use_checkpoint})"
        logger.info(path_file)
        return path_file


_restarter = Restarter()

create_parser = _restarter.create_parser


def main():
    _restarter.restart()


if "sphinx" in sys.modules:
    from textwrap import indent
    from unittest.mock import patch

    with patch.object(sys, "argv", ["snek-restart"]):
        parser = create_parser()

    __doc__ += """
Help message
------------

.. code-block::

""" + indent(
        parser.format_help(), "    "
    )



import pandas as pd
import pytest


@pytest.mark.slow
def test_tgv(sim_tgv_executed):
    sim = sim_tgv_executed

    df = sim.output.spatial_means.load()
    assert isinstance(df, pd.DataFrame)
    assert len(df) > 2
    sim.output.spatial_means.plot()



import os
import shutil
from collections import defaultdict

import pytest

from snek5000.util.smake import (
    append_debug_flags,
    ensure_env,
    set_compiler_verbosity,
)


def test_ensure_env():
    # Remove Nek5000/bin from PATH for the sake of the test
    os.environ["PATH"] = ":".join(
        path for path in os.getenv("PATH").split(":") if "Nek5000/bin" not in path
    )
    assert shutil.which("makenek") is None, "Nek5000/bin is globally available?!"

    ensure_env()
    assert shutil.which("makenek"), "Nek5000/bin is not included in the PATH"


@pytest.mark.parametrize("verbosity", (0, 1, 2))
def test_verbosity(verbosity):
    config = defaultdict(str)
    set_compiler_verbosity(config, verbosity)
    if not verbosity:
        assert " -w " in config["FFLAGS"]
        assert " -w " in config["CFLAGS"]
    elif verbosity == 1:
        assert " -w " not in config["FFLAGS"]
        assert " -w " not in config["CFLAGS"]
        assert " -Wall " not in config["FFLAGS"]
        assert " -Wall " not in config["CFLAGS"]
    elif verbosity == 2:
        assert " -Wall " in config["FFLAGS"]
        assert " -Wall " in config["CFLAGS"]


@pytest.mark.parametrize("warnings", (True, False))
def test_debug_flags(warnings):
    debug_state = os.getenv("SNEK_DEBUG", "")
    os.environ["SNEK_DEBUG"] = "1"

    config = defaultdict(str)
    append_debug_flags(config)

    assert all("-O0 -g" in config[k] for k in ("CFLAGS", "FFLAGS"))

    os.environ["SNEK_DEBUG"] = debug_state



__common_css_variables = {
    "font-size--small--2": "84%",
}
light_css_variables = {
    "color-brand-primary": "#1954a6",
    "color-brand-content": "#1954a6",
    **__common_css_variables,
}
dark_css_variables = {**__common_css_variables}

#  # Material theme options (see theme.conf for more information)
#  html_theme_options = {
#      # Set the name of the project to appear in the navigation.
#      "nav_title": f"{project} documentation",
#      # ??
#      "touch_icon": html_favicon,
#      # Logo to the left of the title
#      "logo_icon": "&#x1F30A",  # https://emojipedia.org/water-wave/
#      #
#      "nav_links": [{"href": "doxygen/index", "title": "fortran docs", "internal": True}],
#      # Specify a base_url used to generate sitemap.xml. If not
#      # specified, then no sitemap will be built.
#      "base_url": f"https://{project}.readthedocs.io",
#      # Set the color and the accent color
#      "color_primary": "indigo",
#      "color_accent": "deep-purple",
#      # Set the repo location to get a badge with stats
#      "repo_url": f"https://github.com/snek5000/{project}/",
#      "repo_name": project,
#      # Visible levels of the global TOC; -1 means unlimited
#      "globaltoc_depth": 2,
#      # If False, expand all TOC entries
#      "globaltoc_collapse": True,
#      # If True, show hidden TOC entries
#      "globaltoc_includehidden": False,
#  }



"""Interface for Nek5000 history points

"""
from io import StringIO

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from snek5000.util import repeat

INDEX_USERPARAM_HISTORY_POINTS = 10


class HistoryPoints:
    """History points (Nek5000 output)

    See https://nek5000.github.io/NekDoc/problem_setup/features.html#history-points

    """

    def __init__(self, output=None):
        self.output = output
        sim = output.sim
        params = sim.params
        hp_params = params.output.history_points

        self.coords = hp_params.coords

        if self.coords is None:
            return

        self.nb_points = len(self.coords)
        self.coords = np.array(self.coords)
        shape = self.coords.shape

        self._data = None

        if shape[0] >= params.oper.max.hist or shape[1] != params.oper.dim:
            raise ValueError(
                "Shape of params.output.history_points.coords are not compatible."
                f"Should be (< {params.oper.max.hist = }, == {params.oper.dim = })"
            )

        self.path_file = output.path_session / f"{sim.info_solver.short_name}.his"
        if self.path_file.exists() or not self.output._has_to_save:
            return

        # temporary?
        output.path_session.mkdir(exist_ok=True)

        with open(self.path_file, "w") as file:
            file.write(f"{self.nb_points} !number of monitoring coords\n")
            np.savetxt(file, self.coords, fmt="%.7E")

    @classmethod
    def _complete_params_with_default(cls, params):
        params.output._set_child(
            "history_points",
            attribs={"coords": None, "write_interval": 100},
        )
        params.output.history_points._record_nek_user_params(
            {"write_interval": INDEX_USERPARAM_HISTORY_POINTS}
        )
        params.output.history_points._set_doc(
            f"""
User parameter for history_points (subroutine hpts):

- ``coords``: list[tuple[float, float]] for 2D or list[tuple[float, float, float]] for 3D

   Coordinates of probes which gets written into the ``*.his`` file. Serves as input
   array for subroutine `hpts()` in Nek5000.

- ``write_interval``: int

   Interval to invoke `hpts()` subroutine. Gets recorded as user parameter number
   {INDEX_USERPARAM_HISTORY_POINTS}. This should be coupled within `userchk()` subroutine
   as follows:

   .. code-block:: fortran

      integer nit_hist

      nit_hist = abs(UPARAM(10))
      if (lhis.gt.1) then
          if (mod(ISTEP,nit_hist).eq.0) then
              call hpts()
          endif
      endif
"""
        )

    def load(self):
        if self._data is None:
            return self._load_full()
        else:
            nb_points = self.coords.shape[0]
            nb_data_lines_read = len(self._data)

            size_file = self.path_file.stat().st_size

            with open(self.path_file) as file:
                first_line = file.readline()
                line_coord = file.readline()
                for _ in repeat(nb_points - 1):
                    file.readline()

                line_data = file.readline()

            nb_chars_read = (
                len(first_line)
                + nb_points * len(line_coord)
                + nb_data_lines_read * len(line_data)
            )
            nb_chars_not_read = size_file - nb_chars_read
            if nb_chars_not_read == 0:
                return self.coords, self._data

            with open(self.path_file) as file:
                file.seek(nb_chars_read)
                lines = file.readlines()

            df_more = self._create_df_from_lines(lines, nb_points)
            if len(df_more) > 0:
                self._data = pd.concat(
                    [self._data, df_more], ignore_index=True, sort=False
                )

            return self.coords, self._data

    def _load_full(self):

        with open(self.path_file) as file:
            nb_points = int(file.readline().split(" ", 1)[0])
            coords = np.loadtxt(file, max_rows=nb_points)
            lines = file.readlines()

        if coords.ndim == 1:
            coords = coords.reshape((1, coords.size))

        # ('x', 'y') or ('x', 'y', 'z')
        columns = tuple("xyz"[: self.output.sim.params.oper.dim])
        self.coords = pd.DataFrame(coords, columns=columns)

        df = self._create_df_from_lines(lines, nb_points)
        self._data = df

        return self.coords, df

    def _create_df_from_lines(self, lines, nb_points):

        columns = ["time", "ux", "uy"]

        sim = self.output.sim
        if sim.params.oper.dim == 3:
            columns.append("uz")

        columns.append("pressure")

        for key in ("temperature", "scalar01"):
            if key not in sim.info_solver.par_sections_disabled:
                columns.append(key)

        # we want to be able to load data during the simulation
        if lines:
            nb_numbers_per_line = len(lines[0])

            while len(lines[-1]) != nb_numbers_per_line:
                lines = lines[:-1]

            nb_times = len(lines) // nb_points
            lines = lines[: nb_times * nb_points]
            if lines:
                df = pd.read_fwf(StringIO("\n".join(lines)), header=None)
                df.columns = columns
            else:
                df = pd.DataFrame({}, columns=columns)
        else:
            nb_times = 0
            df = pd.DataFrame({}, columns=columns)

        index_points = list(range(nb_points)) * nb_times
        df["index_points"] = index_points

        return df

    def plot(self, key="ux"):
        coords, df = self.load()

        fig, ax = plt.subplots()

        for index in range(self.nb_points):
            df_point = df[df.index_points == index]
            signal = df_point[key]
            times = df_point["time"]
            ax.plot(times, signal, label=str(tuple(coords.iloc[index])))

        ax.set(xlabel="time", ylabel=key, title=self.output.summary_simul)
        fig.legend()
        fig.tight_layout()
        return ax

    def load_1point(self, index_point, key=None):
        coords, df = self.load()
        df_point = df[df.index_points == index_point]
        if key is not None:
            df_point = df_point[[key, "time"]]
        return coords, df_point

    def plot_1point(self, index_point, key, tmin=None, tmax=None):
        coords, df_point = self.load_1point(index_point, key)
        fig, ax = plt.subplots()

        signal = df_point[key]
        times = df_point["time"]

        if tmin is not None:
            signal = signal[times > tmin]
            times = times[times > tmin]

        if tmax is not None:
            signal = signal[times < tmax]
            times = times[times < tmax]

        ax.plot(times, signal)
        tmp = ", ".join("xyz"[: self.output.sim.params.oper.dim])
        ax.set(
            xlabel="time",
            ylabel=key,
            title=(
                self.output.summary_simul
                + f"\n{key}, ({tmp}) = {tuple(coords.values[index_point])}"
            ),
        )

        fig.tight_layout()
        return ax



"""Snakemake helper utilities"""
import os


def ensure_env():
    """Ensure environment variables ``NEK_SOURCE_ROOT`` and ``PATH`` are
    intact.

    """
    from .. import get_nek_source_root

    NEK_SOURCE_ROOT = get_nek_source_root()

    if not os.getenv("NEK_SOURCE_ROOT"):
        os.environ["NEK_SOURCE_ROOT"] = NEK_SOURCE_ROOT

    if NEK_SOURCE_ROOT not in os.getenv("PATH"):
        os.environ["PATH"] = ":".join([NEK_SOURCE_ROOT + "/bin", os.getenv("PATH")])


def set_compiler_verbosity(config, verbosity):
    """Set Fortran compiler warnings verbosity:

    Parameters
    ----------
    config: dict
    verbosity: int
        - 0 == suppress warnings
        - 1 == do nothing
        - 2 == all warnings

    """
    flags = {0: " -w ", 1: "", 2: " -Wall "}
    warning_flag = flags[int(verbosity)]
    config["FFLAGS"] = config.get("FFLAGS", "") + warning_flag
    config["CFLAGS"] = config.get("CFLAGS", "") + warning_flag


def append_debug_flags(config):
    """Append to commonly used gcc / gfortran debug flags if ``SNEK_DEBUG``
    environment is activated.

    Parameters
    ----------
    config: dict
        Snakemake configuration. Should contain ``CFLAGS`` and ``FFLAGS`` keys.

    """
    if os.getenv("SNEK_DEBUG"):
        config["CFLAGS"] = config.get("CFLAGS", "") + " -O0 -g"
        config["FFLAGS"] = (
            config.get("FFLAGS", "")
            + " -O0 -g -ffpe-trap=invalid,zero,overflow -DDEBUG "
        )



"""KTH base solver
==================

"""
from pathlib import Path

# NOTE: Ideally InfoSolverMake should be used. Here, this won't work because
# user files are missing.
# from ..info import InfoSolverMake as _InfoSolver
from .. import logger
from ..info import InfoSolverNek as _InfoSolver
from ..util import docstring_params
from .base import SimulNek


class InfoSolverKTH(_InfoSolver):
    """Contain the information on a :class:`snek5000.solvers.kth.SimulKTH`
    instance.

    """

    def _init_root(self):
        super()._init_root()
        self.module_name = "snek5000.solvers.kth"
        self.class_name = "Simul"
        self.short_name = "kth"


class SimulKTH(SimulNek):
    """A base class which incorporates parameters for KTH toolbox also.

    Examples
    --------

    >>> from snek5000.solvers.kth import Simul
    >>> params = Simul.create_default_params()
    >>> sim = Simul(params)

    """

    InfoSolver = InfoSolverKTH

    @classmethod
    def _complete_params_with_default(cls, params):
        params = super()._complete_params_with_default(params)

        for section in ("runpar", "monitor", "chkpoint", "stat"):
            params.nek._set_child(section)

        attribs = {"parf_write": False, "parf_name": "outparfile"}
        params.nek.runpar._set_attribs(attribs)
        params.nek.runpar._set_doc(
            """
    Runtime parameter section for rprm module (`KTH toolbox <https://github.com/KTH-Nek5000/KTH_Toolbox>`__)

    - ``parf_write``: Do we write runtime parameter file
    - ``parf_name``: Runtime parameter file name for output (without .par)
"""
        )

        attribs = {"log_level": 4, "wall_time": "23:45"}
        params.nek.monitor._set_attribs(attribs)
        params.nek.monitor._set_doc(
            """
    Runtime parameter section for monitor module (`KTH toolbox <https://github.com/KTH-Nek5000/KTH_Toolbox>`__)

    - ``log_level``: Logging threshold for toolboxes
    - ``wall_time``: Simulation wall time
"""
        )

        attribs = {"read_chkpt": False, "chkp_fnumber": 1, "chkp_interval": 250}
        params.nek.chkpoint._set_attribs(attribs)
        params.nek.chkpoint._set_doc(
            """
    Runtime parameter section for checkpoint module (`KTH toolbox <https://github.com/KTH-Nek5000/KTH_Toolbox>`__)

    - ``read_chkpt``: Restart from checkpoint
    - ``chkp_fnumber``: Restart file number
    - ``chkp_interval``: Checkpoint saving frequency (number of time steps)
"""
        )

        attribs = {"av_step": 4, "io_step": 50}
        params.nek.stat._set_attribs(attribs)
        params.nek.stat._set_doc(
            """
    Runtime parameter section for statistics module (`KTH toolbox <https://github.com/KTH-Nek5000/KTH_Toolbox>`__)

    - ``av_step``: Frequency of averaging
    - ``io_step``: Frequency of file saving
"""
        )

        # Document all params
        for child_name in ("runpar", "monitor", "chkpoint", "stat"):
            child = getattr(params.nek, child_name)
            child._autodoc_par(indent=4)

        return params

    def create_symlinks_checkpoint_files(self, path_run):
        """Create symlinks towards checkpoint files"""
        paths = sorted(Path(path_run).glob("rs6*"))
        if not paths:
            raise FileNotFoundError("No restart file in {path_run}")
        logger.info("Symlinking checkpoint files")
        for src in paths:
            dest = Path(self.output.path_run) / src.name
            dest.symlink_to(src)


Simul = SimulKTH
Simul.__doc__ += """

    Notes
    -----

    Here, only the documention for ``params.nek`` is displayed. For the
    documentation on ``params.oper`` see :mod:`snek5000.operators`.

""" + docstring_params(
    Simul, indent_len=4
)



"""Runtime parameters
=====================
Scripting interface for Nek5000 :ref:`parameter file <nek:case_files_par>`.

"""
import json
import logging
import os
import sys
import textwrap
from ast import literal_eval
from configparser import ConfigParser
from io import StringIO
from math import nan
from pathlib import Path

from inflection import camelize, underscore

from fluidsim_core.params import Parameters as _Parameters

from .log import logger
from .solvers import get_solver_short_name, import_cls_simul

literal_python2nek = {
    nan: "<real>",
    "nan": "nan",
    "None": "none",
    "True": "yes",
    "False": "no",
}
literal_nek2python = {v: k for k, v in literal_python2nek.items()}
literal_prune = ("<real>", "", "nan")

#: JSON file name to which recorded user_params are saved
filename_map_user_params = "map_user_params.json"


def _as_nek_value(input_value):
    """Convert Python values to equivalent Nek5000 par values."""
    # Convert to string to avoid hash collisions
    # hash(1) == hash(True)
    literal = str(input_value) if input_value is not nan else nan
    value = literal_python2nek.get(literal, input_value)
    return value


def camelcase(value):
    """Convert strings to ``camelCase``."""
    return camelize(str(value).lower(), uppercase_first_letter=False)


def _check_user_param_index(idx):
    """Check if the index of user parameter is within bounds"""
    if idx > 20:
        raise ValueError(f"userParam {idx = } > 20")


def _as_python_value(input_value):
    """Convert Nek5000 par values to equivalent Python values if possible."""
    value = literal_nek2python.get(str(input_value), input_value)

    try:
        return literal_eval(value)
    except (SyntaxError, ValueError):
        return value


def load_params(path_dir="."):
    """Load a :class:`snek5000.params.Parameters` instance from `path_dir`.

    Parameters
    ----------
    path_dir : str or path-like
        Path to a simulation directory.

    Returns
    -------
    params: :class:`snek5000.params.Parameters`

    """
    from snek5000.util.files import _path_try_from_fluidsim_path

    path_dir = _path_try_from_fluidsim_path(path_dir)
    short_name = get_solver_short_name(path_dir)
    Simul = import_cls_simul(short_name)

    return Simul.load_params_from_file(
        path_xml=path_dir / "params_simul.xml", path_par=path_dir / f"{short_name}.par"
    )


class Parameters(_Parameters):
    """Container for reading, modifying and writing :ref:`par files
    <nek:case_files_par>`.

    :param tag: A string representing name of case files (for example: provide
                 ``"abl"`` for files like ``abl.usr, abl.par`` etc).

    """

    @classmethod
    def _load_params_simul(cls, path=None):
        """Alias for :func:`load_params`"""
        return load_params(path or Path.cwd())

    def __init__(self, *args, **kwargs):
        comments = ("#",)
        self._set_internal_attr(
            "_par_file",
            ConfigParser(comment_prefixes=comments, inline_comment_prefixes=comments),
        )
        # Only enabled parameters would be written into par file
        self._set_internal_attr("_enabled", True)
        # User parameters sections should begin with an underscore
        self._set_internal_attr("_user", True)

        super().__init__(*args, **kwargs)

        # Like in Python Nek5000's par files are case insensitive.
        # However for consistency, case sensitivity is enforced:
        self._par_file.optionxform = str

    def _make_dict_attribs(self):
        d = super()._make_dict_attribs()
        # Append internal attributes
        d.update({"_enabled": self._enabled, "_user": self._user})
        if hasattr(self, "_recorded_user_params"):
            d["_recorded_user_params"] = self._recorded_user_params
        return d

    def __update_par_section(
        self, section_name, section_dict, has_to_prune_literals=True
    ):
        """Updates a section of the ``par_file`` object from a dictionary."""
        par = self._par_file

        # Start with underscore if it is a user section
        section_name_par = "_" if section_dict["_user"] else ""
        section_name_par += section_name.upper().lstrip("_")

        if section_name_par not in par.sections():
            par.add_section(section_name_par)

        if "_recorded_user_params" in section_dict:
            recorded_user_params = section_dict.pop("_recorded_user_params")
        else:
            recorded_user_params = False

        for option, value in section_dict.items():
            value = _as_nek_value(value)

            if has_to_prune_literals and value in literal_prune:
                continue

            # Make everything consistent where values refer to option names
            # if option in ("stop_at", "write_control"):
            if str(value) in section_dict:
                value = camelcase(value)

            par.set(section_name_par, camelcase(option), str(value))

        # _recorded_user_params -> userParam%%
        if not recorded_user_params:
            return
        params = self._parent
        if self._tag != "nek" or params._tag != "params":
            raise RuntimeError(
                "_recorded_user_params should only be in params.nek.general"
            )
        for idx_uparam in sorted(recorded_user_params.keys()):
            tag = recorded_user_params[idx_uparam]
            _check_user_param_index(idx_uparam)
            value = _as_nek_value(params[tag])
            par.set(
                section_name_par,
                f"userParam{idx_uparam:02d}",
                str(value),
            )

    def _sync_par(self, has_to_prune_literals=True, keep_all_sections=False):
        """Sync values in param children and attributes to ``self._par_file``
        object.

        """
        if self._tag_children:
            data = [
                (child, getattr(self, child)._make_dict_tree())
                for child in self._tag_children
            ]
        else:
            # No children
            data = [(self._tag, self._make_dict_attribs())]

        for child, d in data:
            # Section name is often written in [UPPERCASE]
            section_name = child.upper()
            self.__update_par_section(
                section_name, d, has_to_prune_literals=has_to_prune_literals
            )

        self.__tidy_par(keep_all_sections)

    def __tidy_par(self, keep_all_sections=False):
        """Remove internal attributes and disabled sections from par file."""
        par = self._par_file
        for section_name in par.sections():
            par.remove_option(section_name, "_user")

            if keep_all_sections:
                enabled = True
            else:
                enabled = par.getboolean(section_name, "_enabled")
            if enabled:
                par.remove_option(section_name, "_enabled")
            else:
                par.remove_section(section_name)

    def _autodoc_par(self, indent=0):
        """Autodoc a code block with ``ini`` syntax and set docstring."""
        self._sync_par(has_to_prune_literals=False, keep_all_sections=True)
        docstring = "\n.. code-block:: ini\n\n"
        with StringIO() as output:
            self._par_file.write(output)
            ini = output.getvalue()

        docstring += textwrap.indent(ini, "   ")

        if ini:
            self._set_doc(self._doc + textwrap.indent(docstring, " " * indent))

    def _record_nek_user_params(self, nek_params_keys, overwrite=False):
        """Record some Nek user parameters

        Examples
        --------

        >>> params._record_nek_user_params({"prandtl": 2, "rayleigh": 3})
        >>> params.output.history_points._record_nek_user_params({"write_interval": 4})

        This is going to set or modify the internal attribute
        ``params.nek.general._recorded_user_params`` to ``{2: "prandtl", 3:
        "rayleigh", 4: "output.other.write_interval"}``.

        This attribute is then used to write the ``[GENERAL]`` section of the
        .par file.

        Note that this attribute is only for ``params.nek.general`` and should
        never be set for other parameter children.
        """
        # we need to find where is self in the tree compared to `params`
        current = self
        parent = current._parent
        tag = current._tag
        path = tag

        # iterate up the `params` tree to the top
        while not (parent is None and tag == "params") and not (
            parent._tag == "info_simul" and tag == "params"
        ):
            current = parent
            parent = current._parent
            tag = current._tag
            path = f"{tag}.{path}"

        params = current
        assert params._tag == "params"

        # path relative to params:
        # we have `(path, name)` equal to
        # `("params.output.history_points", "write_interval")` or
        # `("params", "rayleigh")` and we want to end up with
        # `"output.history_points.write_interval"` or `rayleigh`, resp.
        path = path[len("params") :]
        if path.startswith("."):
            path = path[1:]
        if path:
            path = path + "."

        user_params = {}
        for name, key in nek_params_keys.items():
            user_params[key] = f"{path}{name}"

        # Useful while building isolated `params` for a specific class,
        # for e.g.: Operators, Output etc.
        if not hasattr(params, "nek"):
            log_level = logging.DEBUG if "sphinx" in sys.modules else logging.WARNING
            logger.log(
                log_level,
                (
                    "Attribute params.nek does not exist, skipping "
                    "initializing user parameters."
                ),
            )
            return

        general = params.nek.general
        if not hasattr(general, "_recorded_user_params"):
            general._set_internal_attr("_recorded_user_params", {})

        if overwrite:
            general._recorded_user_params.update(user_params)
            return

        for key, value in user_params.items():
            if key in general._recorded_user_params:
                raise ValueError(
                    f"{key = } already used for user parameter "
                    f"{general._recorded_user_params[key]}"
                )
            general._recorded_user_params[key] = value

    def _change_index_userparams(self, user_params):
        """Change indices for user parameters

        This method can be used in the ``create_default_params`` class method
        of a solver to overwrite the default indices used in the base snek5000
        package.

        This method checks that no already recorded parameters are overwritten.
        To overwrite a parameter, use ``_record_nek_user_params`` with the
        ``overwrite`` argument.

        Examples
        --------

        >>> params._change_index_userparams({8: "output.history_points.write_interval"}

        """

        if self._tag != "params":
            raise ValueError(
                "The method `_change_index_userparams` has to be called "
                "directly with the root `params` object."
            )

        try:
            general = self.nek.general
        except AttributeError:
            raise AttributeError("No `params.nek.general` attribute.")

        try:
            recorded_user_params = general._recorded_user_params
        except AttributeError:
            raise AttributeError(
                "No `general._recorded_user_params` attribute. This attribute "
                "can be created with `_record_nek_user_params`."
            )

        # check that no user parameters are overwritten
        modified_labels = []
        for index in user_params:
            try:
                modified_labels.append(recorded_user_params[index])
            except KeyError:
                pass
        values = user_params.values()
        for label in modified_labels:
            if label not in values:
                raise ValueError(
                    f"The value {label} would be removed from the user params."
                )

        reverted = {value: key for key, value in recorded_user_params.items()}
        for label in user_params.values():
            try:
                key = reverted[label]
            except KeyError:
                raise ValueError(
                    f"User parameter {label = } is not already recorded. "
                    "Use `_record_nek_user_params`"
                )
            del recorded_user_params[key]

        recorded_user_params.update(user_params)

    def _save_as_xml(self, path_file=None, comment=None, find_new_name=False):
        """Invoke :func:`_save_recorded_user_params` and then save to an XML file at ``path_file``."""
        try:
            user_params = self.nek.general._recorded_user_params
        except AttributeError:
            pass
        else:
            if path_file is None:
                path_dir = Path.cwd()
            else:
                path_dir = Path(path_file).parent
            _save_recorded_user_params(user_params, path_dir)

        return super()._save_as_xml(
            path_file=path_file, comment=comment, find_new_name=find_new_name
        )


def _save_recorded_user_params(user_params, path_dir):
    """Save a JSON file from a dictionary denoting ``user_params``"""
    with open(path_dir / filename_map_user_params, "w") as file:
        json.dump(user_params, file)


def _load_recorded_user_params(path):
    """Load a JSON file and return a dictionary denoting ``user_params``"""
    with open(path) as file:
        tmp = json.load(file)
    return {int(key): value for key, value in tmp.items()}


def _check_path_like(path):
    """Ensure input is a path-like object"""
    if not isinstance(path, os.PathLike):
        raise TypeError(f"Expected path-like object, not {type(path) = }")


def _get_params_nek(params):
    """Check if params is the top level object (via the ``_tag`` attribute) and
    return the ``params.nek`` object.

    Parameters
    ----------
    params: :class:`Parameters`
        The ``params`` object

    Returns
    -------
    params.nek: :class:`Parameters`
        The ``params.nek`` object

    """
    if not isinstance(params, Parameters):
        raise TypeError
    if params._tag != "params":
        raise ValueError(f'{params._tag = } != "params"')

    if params.nek._tag != "nek":
        raise RuntimeError(f'{params.nek._tag =} != "nek"')
    return params.nek


def _save_par_file(params, path, mode="w"):
    """Save the ``params.nek`` object as a `.par` file."""
    nek = _get_params_nek(params)
    nek._sync_par()

    _check_path_like(path)
    with open(path, mode) as fp:
        nek._par_file.write(fp)

    if hasattr(nek.general, "_recorded_user_params"):
        _save_recorded_user_params(nek.general._recorded_user_params, path.parent)


def _str_par_file(params):
    """Preview contents of the resulting `.par` file as a string"""
    nek = _get_params_nek(params)
    nek._sync_par()

    with StringIO() as output:
        nek._par_file.write(output)
        return output.getvalue()


def complete_params_from_par_file(params, path):
    """Populate the ``params.nek`` object by reading a `.par` file and
    :attr:`filename_map_user_params`.

    """
    _check_path_like(path)
    if not path.exists():
        raise IOError(f"{path} does not exist.")

    nek = _get_params_nek(params)
    nek._par_file.read(path)

    recorded_user_params_path = path.with_name(filename_map_user_params)
    if recorded_user_params_path.exists():
        recorded_user_params = _load_recorded_user_params(recorded_user_params_path)
    elif hasattr(nek.general, "_recorded_user_params"):
        recorded_user_params = nek.general._recorded_user_params
    else:
        recorded_user_params = {}

    for section in nek._par_file.sections():
        params_child = getattr(nek, section.lower().lstrip("_"))

        for option, value in nek._par_file.items(section):
            value = _as_python_value(value)

            # userParam%% -> user_params
            if option.lower().startswith("userparam"):
                idx_uparam = int(option[-2:])
                _check_user_param_index(idx_uparam)
                if idx_uparam not in recorded_user_params:
                    logger.warning(
                        f"{idx_uparam = } not in {recorded_user_params = } so we"
                        "cannot update the right parameter in the object `params`."
                        "It might be because you load a simulation done with "
                        "an old snek5000 (< 0.8), or it might be a bug :-)"
                    )
                tag = recorded_user_params[idx_uparam]
                # set the corresponding parameter
                params[tag] = value
            else:
                attrib = underscore(option)
                setattr(params_child, attrib, value)


def _complete_params_from_xml_file(params, path_xml):
    """Populate the ``params.nek`` object by reading a `.xml` file and
    :attr:`filename_map_user_params`.

    """
    _check_path_like(path_xml)

    params._load_from_xml_file(str(path_xml))
    nek = _get_params_nek(params)
    path_recorded_user_params = Path(path_xml).parent / filename_map_user_params
    if path_recorded_user_params.exists():
        nek.general._set_internal_attr(
            "_recorded_user_params",
            _load_recorded_user_params(path_recorded_user_params),
        )


create_params = Parameters._create_params



"""Post simulation archive-creation utilities"""
import shlex
import shutil
import subprocess
from glob import iglob
from pathlib import Path
from shutil import rmtree

from .. import logger
from . import modification_date
from .files import next_path


def archive(tarball, items=(), remove=False, readonly=False):
    """Archive simulation contents into a tarball / compress a tarball.

    Examples
    --------
    >>> archive("test.tar.xz", ["file1", "file2"])
    # runs tar -cvf test.tar.xz file1 file2

    >>> archive("test.tar.gz")
    # runs gzip test.tar

    """
    logger.info(repr(locals()))
    if not tarball:
        raise IOError(
            "Abort archiving because of illegal tarball name / empty input! "
            f"tarball: {tarball}, items: {items}"
        )

    # prepare output directory
    dest = Path.cwd() / "data"
    dest.mkdir(exist_ok=True)

    if items:
        cmd_output, output = exec_tar(tarball, items, remove)
    else:
        cmd_output, output = exec_compress(tarball)

    if readonly:
        output.chmod(0o444)

    return cmd_output


def clean_simul(case, tarball):
    """Clean a simulation after archiving into a tarball."""
    if tarball and not Path(tarball).exists():
        raise IOError(
            f"Archive {tarball} not found. Refusing to clean "
            "simulation files! Run 'snakemake archive'"
        )

    remove(Path(f"{case}.{ext}") for ext in "re2 ma2 log f nek5000".split())
    remove(
        Path(file)
        for file in "makefile box.tmp compiler.out SESSION.NAME GIT_REVISION.txt nek5000".split()
    )
    remove([Path("obj")])
    field_files = Path.cwd().glob(f"*{case}?.f?????")
    remove(field_files)


def exec_compress(tarball):
    tar, tarball, compress_format = parse_args_from_filename(tarball)

    if not compress_format:
        logger.error(f"Cannot compress: {tar} -> {tarball}")
        return "", tarball

    # tarball will be compressed in place when no items are provided
    main = compress_cmd(compress_format).split()
    output = next_path(tarball)

    if str(output.absolute()) != str(tarball.absolute()):
        if compress_format == ".zst":
            kwargs = ("-o", output)
            cmd = [*main, str(tar), *kwargs]
            cmd_output = subprocess.check_output(cmd)
        else:
            # do something like gzip -c file.tar > file.tar.gz using
            main.append("--stdout")
            with open(tarball, "w") as fp:
                cmd = [*main, str(tar)]
                subprocess.run(cmd, stdout=fp)

            cmd_output = ""
    else:
        cmd = [*main, str(tar)]
        cmd_output = subprocess.check_output(cmd)

    return cmd_output, output


def exec_tar(tarball, items, remove):
    tar, tarball, compress_format = parse_args_from_filename(tarball)

    if compress_format:
        output = tarball
        if tar.exists():
            raise IOError("Cannot append file items into a compressed archive!")
    else:
        output = tar

    main = shlex.split(tar_cmd(compress_format, remove=remove, append=output.exists()))

    # run command
    items = [str(i) for i in items]
    cmd = [*main, str(tar), *items]
    cmd_output = subprocess.check_output(cmd)
    return cmd_output, output


def parse_args_from_filename(tarball):
    tarball = Path(str(tarball))

    # split archive.tar.zst -> archive, .tar, .zst
    name, dottar, compress_format = tarball.name.partition(".tar")
    tar = tarball.parent / (name + dottar)
    return tar, tarball, compress_format


def compress_cmd(compress_format):
    compress_program = {
        ".gz": "pigz" if shutil.which("pigz") else "gzip",
        ".xz": "xz",
        ".lz4": "lz4",
        ".zst": "zstdmt --rm",
    }
    cmd = compress_program[compress_format]
    return cmd


def tar_cmd(compress_format="", remove=False, append=False):
    archive_program = {
        "": "tar",
        ".gz": f"tar --use-compress-program {compress_cmd('.gz')}",
        ".xz": "tar",
        ".lz4": "tar --use-compress-program lz4",
        ".zst": "tar --use-compress-program zstdmt",
    }
    # For benchmarking ...
    # cmd = "time " +
    cmd = archive_program[compress_format]
    if remove:
        cmd += " --remove-files "
    elif shutil.which("bsdtar"):
        # Use bsdtar if available
        # bsdtar is faster, but does not support --remove-files option
        cmd = "bsdtar"

    if append:
        return cmd + " --append --file"
    else:
        # create new tarball
        return cmd + " -cvf"


def tar_name(
    root_name="abl",
    pattern="*.f?????",
    compress_format="",
    subdir="data",
    default_prefix="test",
):
    """Generate a tarball name based on contents of current working
    directory.

    """
    modified_dates = [modification_date(f) for f in iglob(pattern)]
    cwd = Path.cwd().name
    if modified_dates:
        if cwd == root_name:
            timestamp = max(modified_dates)
            basename = f"{default_prefix}-{timestamp}"
        else:
            basename = cwd

        return f"{subdir}/{basename}.tar{compress_format}"
    else:
        return ""


def remove(items):
    """Equivalent to rm -rf"""
    print("Removing", items)
    for item in items:
        if item.is_dir():
            rmtree(item, ignore_errors=True)
        else:
            try:
                item.unlink(missing_ok=True)
            except TypeError:
                # python < 3.8
                pass



from snek5000.output.base import Output as OutputBase
from snek5000.resources import get_base_templates

box, makefile_usr, size = get_base_templates()


class OutputTGV(OutputBase):
    template_box = box
    template_makefile_usr = makefile_usr
    template_size = size

    @property
    def makefile_usr_sources(self):
        """
        Sources for inclusion to makefile_usr.inc
        Dict[directory]  -> list of source files
        """
        return {}

    @classmethod
    def _set_info_solver_classes(cls, classes):
        """Set the the classes for info_solver.classes.Output"""
        super()._set_info_solver_classes(classes)

        classes._set_child(
            "SpatialMeans",
            dict(
                module_name="snek5000.output.spatial_means",
                class_name="SpatialMeans",
            ),
        )


Output = OutputTGV



"""Task runner for the developer

Usage
-----

   nox -l

   nox -s <session>

   nox -k <keyword>
or:

   make <session>

execute ``make list-sessions```` or ``nox -l`` for a list of sessions.

"""
import os
import re
import shlex
import shutil
from functools import partial
from pathlib import Path

import nox

PACKAGE = "snek5000"
CWD = Path.cwd()
if (CWD / "poetry.lock").exists():
    BUILD_SYSTEM = "poetry"
    PACKAGE_SPEC = "pyproject.toml"
else:
    BUILD_SYSTEM = "setuptools"
    PACKAGE_SPEC = "setup.cfg"

try:
    NEK_SOURCE_ROOT = os.environ["NEK_SOURCE_ROOT"]
except KeyError:
    raise RuntimeError(
        "Environment variable NEK_SOURCE_ROOT should be set "
        "pointing to Nek5000 top level directory."
    )

TEST_ENV_VARS = {
    "NEK_SOURCE_ROOT": NEK_SOURCE_ROOT,
    "SNEK_DEBUG": "1",
}
if os.getenv("CI"):
    TEST_ENV_VARS["PYTEST_ADDOPTS"] = "--color=yes"

EXTRA_REQUIRES = ("main", "docs", "tests", "dev")

no_venv_session = partial(nox.session, venv_backend="none")
nox.options.sessions = ["tests"]


def run_ext(session, cmd):
    """Run an external command, i.e. outside a nox managed virtual envionment"""
    session.run(*shlex.split(cmd), external=True)


def rmdir(path_dir: str):
    if Path(path_dir).exists():
        shutil.rmtree(path_dir)


def poetry_install(session, *args):
    """Install with dependencies pinned in pyproject.toml"""
    run_ext(session, "python -m poetry install " + " ".join(args))


def pip_install(session, filename, *args):
    """Install with dependencies pinned in requirements/*.txt"""
    run_ext(
        session,
        f"python -m pip install -r requirements/{filename}.txt " + " ".join(args),
    )


def pip_sync(session, filename):
    """Reset developer environment with dependencies pinned in requirements/dev.txt"""
    run_ext(session, f"python -m piptools sync requirements/{filename}.txt")


@no_venv_session
def install(session):
    """Install package."""
    if BUILD_SYSTEM == "poetry":
        poetry_install(session)
    else:
        pip_install(session, "main", ".")


@no_venv_session
def develop(session):
    """Install developer environment."""
    if BUILD_SYSTEM == "poetry":
        poetry_install(session, "--with=dev")
    else:
        pip_install(session, "dev")


@no_venv_session
def sync(session):
    """Sync developer environment."""
    if BUILD_SYSTEM == "poetry":
        poetry_install(session, "--sync", "--with=dev")
    else:
        pip_sync(session, "dev")


@no_venv_session
def requires(session):
    """Pin dependencies"""
    if BUILD_SYSTEM == "poetry":
        run_ext(session, "python -m poetry lock --no-update")
    else:
        session.notify("pip-compile")


@nox.session(name="pip-compile", reuse_venv=True)
@nox.parametrize("extra", [nox.param(extra, id=extra) for extra in EXTRA_REQUIRES])
def pip_compile(session, extra):
    """Pin dependencies to requirements/*.txt

    How to run all in parallel::

        pipx install nox
        make -j requirements

    or::

        nox -l | awk '/pip-compile/{print $2}' | xargs -P6 -I_ nox -s _

    """
    session.install("pip-tools")
    req = Path("requirements")

    if extra == "main":
        in_extra = ""
        in_file = ""
    else:
        in_extra = f"--extra {extra}"
        in_file = req / "vcs_packages.in"

    out_file = req / f"{extra}.txt"

    session.run(
        *shlex.split(
            "python -m piptools compile --resolver backtracking --quiet "
            f"{in_extra} {in_file} {PACKAGE_SPEC} "
            f"-o {out_file}"
        ),
        *session.posargs,
    )

    session.log(f"Removing absolute paths from {out_file}")
    packages = out_file.read_text()
    rel_path_packages = packages.replace("file://" + str(Path.cwd().resolve()), ".")

    # special cases for snek5000-canonical and snek5000-tgv
    for package in ("snek5000-canonical", "snek5000-tgv"):
        rel_path_packages = rel_path_packages.replace(f"{package} @ ", "")

    if extra == "tests":
        tests_editable = out_file.parent / out_file.name.replace(
            "tests", "tests-editable"
        )
        session.log(f"Copying {out_file} with -e flag in {tests_editable}")
        tests_editable.write_text(rel_path_packages)
        session.log(f"Removing -e flag in {out_file}")
        rel_path_packages = re.sub(r"^-e\ \.", ".", rel_path_packages, flags=re.M)

    session.log(f"Writing {out_file}")
    out_file.write_text(rel_path_packages)


def install_with_tests(session, args=()):
    if BUILD_SYSTEM == "poetry":
        session.install("poetry")
        session.run("python", "-m", "poetry", "install", "--with=tests", *args)
        session.run("python", "-m", "poetry", "env", "info")
        return "python", "-m", "poetry", "run", "pytest"
    else:
        session.install("-r", "requirements/tests.txt", *args)
        return "python", "-m", "pytest"


@nox.session
def tests(session):
    """Execute unit-tests using pytest"""
    pytest_cmd = install_with_tests(session)
    session.run(
        *pytest_cmd,
        *session.posargs,
        env=TEST_ENV_VARS,
    )


@no_venv_session(name="tests-cov")
def tests_cov(session):
    """Execute unit-tests using pytest+pytest-cov"""
    session.notify(
        "tests",
        [
            "--cov",
            "--cov-config=pyproject.toml",
            "--no-cov-on-fail",
            "--cov-report=term-missing",
            *session.posargs,
        ],
    )


@nox.session(name="coverage-html")
def coverage_html(session, nox=False):
    """Generate coverage report in HTML. Requires `tests-cov` session."""
    report = Path.cwd() / ".coverage" / "html" / "index.html"
    session.install("coverage[toml]")
    session.run("coverage", "html")

    print("Code coverage analysis complete. View detailed report:")
    print(f"file://{report}")


@no_venv_session(name="format")
def format_(session):
    """Run pre-commit hooks on all files to set and lint code-format"""
    run_ext(session, "pre-commit install")
    run_ext(session, "pre-commit run --all-files")


@nox.session
def lint(session):
    """Run pre-commit hooks on files which differ in the current branch from origin/HEAD."""
    remote = "origin/HEAD" if not session.posargs else session.posargs[0]
    session.install("pre-commit")
    session.run("pre-commit", "install")
    session.run("pre-commit", "run", "--from-ref", remote, "--to-ref", "HEAD")


def _prepare_docs_session(session):
    session.install("-r", "requirements/docs.txt")
    session.chdir("./docs")

    build_dir = Path.cwd() / "_build"
    source_dir = "."
    output_dir = str(build_dir.resolve() / "html")
    return source_dir, output_dir


@nox.session
def docs(session):
    """Build documentation using Sphinx."""
    source, output = _prepare_docs_session(session)
    session.run(
        "python", "-m", "sphinx", "-b", "html", *session.posargs, source, output
    )  # Same as sphinx-build
    print("Build finished.")
    print(f"file://{output}/index.html")


@nox.session(name="docs-autobuild")
def docs_autobuild(session):
    """Build documentation using sphinx-autobuild."""
    source, output = _prepare_docs_session(session)
    session.run(
        "python",
        "-m",
        "sphinx_autobuild",
        "--watch",
        "../src",
        "--re-ignore",
        r"(_build|generated)\/.*",
        source,
        output,
    )  # Same as sphinx-autobuild
    print("Build finished.")
    print(f"file://{output}/index.html")


@no_venv_session
def ctags(session):
    """Runs universal-ctags to build .tags file"""
    sources = {
        "nek5000": str(Path(NEK_SOURCE_ROOT) / "core"),
        "snek5000": "src/snek5000",
    }
    output = ".tags"
    excludes = " ".join(
        (
            f"--exclude={pattern}"
            for pattern in (
                ".snakemake",
                "__pycache__",
                "obj",
                "logs",
                "*.tar.gz",
                "*.f?????",
            )
        )
    )
    run_ext(
        session, f"ctags -f {output} --language-force=Fortran -R {sources['nek5000']}"
    )
    run_ext(session, f"ctags -f {output} {excludes} --append -R {sources['snek5000']}")


@no_venv_session
def testpypi(session):
    """Release clean, build, upload to TestPyPI"""
    session.notify("release-clean")
    session.notify("release-build")
    session.notify("release-upload", ["--repository", "testpypi"])


@no_venv_session
def pypi(session):
    """Release clean, download from TestPyPI, test, upload to PyPI"""
    session.notify("release-clean")
    # NOTE: parametrizing dist_type ends up in erraneous deduplication of sessions
    # by nox
    for dist_type in ("no-binary", "only-binary"):
        session.notify(f"download-testpypi(dist_type='{dist_type}')")
        session.notify(f"release-tests(dist_type='{dist_type}')")
    session.notify("release-upload", ["--repository", "pypi"])


@nox.session(name="download-testpypi")
@nox.parametrize("dist_type", ["no-binary", "only-binary"])
def download_testpypi(session, dist_type):
    """Download from TestPyPI and run tests"""
    (Path.cwd() / "dist").mkdir(exist_ok=True)
    session.chdir("./dist")

    git_tags = session.run(
        "git", "tag", "--list", "--sort=committerdate", external=True, silent=True
    )
    latest_version = git_tags.splitlines()[-1]
    spec = f"{PACKAGE}=={latest_version}"
    session.run(
        "python",
        "-m",
        "pip",
        "index",
        "versions",
        "--index-url",
        "https://test.pypi.org/simple",
        "--pre",
        PACKAGE,
    )
    session.run(
        "python",
        "-m",
        "pip",
        "download",
        "--index-url",
        "https://test.pypi.org/simple",
        "--extra-index-url",
        "https://pypi.org/simple",
        "--pre",
        "--no-deps",
        f"--{dist_type}",
        PACKAGE,
        spec,
    )


@nox.session(name="release-tests")
@nox.parametrize("dist_type", ["no-binary", "only-binary"])
def release_tests(session, dist_type):
    """Execute test suite with build / downloaded package in ./dist"""
    if dist_type == "only-binary":
        pattern = "*.whl"
    else:
        pattern = "*.tar.gz"

    if BUILD_SYSTEM == "poetry":
        poetry_conf = CWD / "poetry.toml"
        assert (
            not poetry_conf.exists()
        ), "Poetry local configuration exists. Please remove to continue"
        session.install("poetry")
        session.run(
            "python", "-m", "poetry", "config", "--local", "virtualenvs.create", "false"
        )
        pytest_cmd = install_with_tests(session, ["--no-root"])
    else:
        pytest_cmd = install_with_tests(session)

    dist_packages = [str(p) for p in Path("./dist").glob(pattern)]
    session.install(*dist_packages)

    try:
        session.run(
            *pytest_cmd,
            env=TEST_ENV_VARS,
        )
    finally:
        if BUILD_SYSTEM == "poetry":
            poetry_conf.unlink()


@no_venv_session(name="release-clean")
def release_clean(session):
    """Remove build and dist directories"""
    session.log("Removing build and dist")
    rmdir("./build/")
    rmdir("./dist/")


@nox.session(name="release-build")
def release_build(session):
    """Build package into dist."""
    session.install("build")
    session.run("python", "-m", "build")


@nox.session(name="release-upload")
def release_upload(session):
    """Upload dist/* to repository testpypi (default, must be configured in ~/.pypirc).
    Also accepts positional arguments to `twine upload` command.

    """
    session.install("twine")
    session.run("twine", "check", "--strict", "dist/*")
    args = session.posargs

    # See
    # https://pypi.org/help/#apitoken and
    # https://twine.readthedocs.io/en/latest/#environment-variables
    env = {"TWINE_USERNAME": "__token__"}

    test_pypi_token = os.getenv("TEST_PYPI_TOKEN")
    pypi_token = os.getenv("PYPI_TOKEN")
    if "testpypi" in args and test_pypi_token:
        env["TWINE_PASSWORD"] = test_pypi_token
    elif pypi_token:
        env["TWINE_PASSWORD"] = pypi_token

    session.run("twine", "upload", *args, "dist/*", env=env)



import tempfile
from pathlib import Path

import pytest

from snek5000.log import logger
from snek5000.params import (
    Parameters,
    _as_nek_value,
    _as_python_value,
    _save_par_file,
    _str_par_file,
    complete_params_from_par_file,
)
from snek5000.util import init_params


@pytest.mark.parametrize("input_value", (True, False, None, "a string", 1, 3.14, "nan"))
def test_python_value_roundtrip(input_value):
    roundtrip_python_nek = _as_python_value(_as_nek_value(input_value))
    assert input_value == roundtrip_python_nek


@pytest.mark.parametrize(
    ("input_value", "equivalent"),
    zip(("yes", "no", "42", "2.71"), (True, False, 42, 2.71)),
)
def test_python_value_roundtrip_evaluated(input_value, equivalent):
    roundtrip_python_nek = _as_python_value(_as_nek_value(input_value))
    assert input_value != roundtrip_python_nek == equivalent


@pytest.mark.parametrize(
    ("input_value", "equivalent"),
    zip(("True", "False", "42", "2.71"), ("yes", "no", 42, 2.71)),
)
def test_nek_value_roundtrip_evaluated(input_value, equivalent):
    roundtrip_nek_python = _as_nek_value(_as_python_value(input_value))
    assert input_value != roundtrip_nek_python == equivalent


def save_par_file_and_read(params, path):
    _save_par_file(params, path)
    with open(path) as file:
        return file.read()


def test_empty_params():
    Parameters(tag="empty")


def test_simul_params():
    from snek5000.solvers.base import Simul

    tmp_dir = Path(tempfile.mkdtemp("snek5000", __name__))

    # create params different from the default params
    params = Simul.create_default_params()
    params.nek.general.dt = 2.0

    path_par = tmp_dir / "tmp.par"
    txt = save_par_file_and_read(params, path_par)
    assert txt

    params1 = Simul.create_default_params()
    complete_params_from_par_file(params1, path_par)

    assert params1.nek.general.dt == params.nek.general.dt

    txt1 = save_par_file_and_read(params1, tmp_dir / "tmp1.par")

    assert txt == txt1


def test_oper_params(oper):
    from snek5000.operators import Operators

    params = init_params(Operators)
    logger.debug(params.oper.max)
    logger.debug(params.oper.max._doc)
    logger.debug(params.oper.elem)
    logger.debug(params.oper.elem._doc)


def test_par_xml_match():
    from phill.solver import Simul

    params = Simul.create_default_params()

    tmp_dir = Path(tempfile.mkdtemp("snek5000", __name__))

    par1 = save_par_file_and_read(params, tmp_dir / "tmp1.par")

    params_xml = Path(params._save_as_xml(str(tmp_dir / "params_simul.xml")))

    try:
        from snek5000.params import Parameters

        nparams = Parameters(tag="params", path_file=params_xml)
    except ValueError:
        # NOTE: used to raise an error, now testing experimentally
        pass
    #  else:
    #      raise ValueError("Parameters(path_file=...) worked unexpectedly.")

    nparams = Simul.load_params_from_file(path_xml=params_xml)

    par2 = save_par_file_and_read(nparams, tmp_dir / "tmp2.par")

    def format_sections(params):
        par = params.nek._par_file

        # no options in the section
        for section_name in par.sections():
            if not par.options(section_name):
                par.remove_section(section_name)

        return sorted(par.sections())

    assert format_sections(params) == format_sections(nparams)

    def format_par(text):
        """Sort non-blank lines"""
        from ast import literal_eval

        ftext = []
        for line in text.splitlines():
            # not blank
            if line:
                # Uniform format for literals
                if " = " in line:
                    key, value = line.split(" = ")
                    try:
                        line = " = ".join([key, str(literal_eval(value))])
                    except (SyntaxError, ValueError):
                        pass

                ftext.append(line)

        return sorted(ftext)

    assert format_par(par1) == format_par(par2)


def test_user_params():

    tmp_dir = Path(tempfile.mkdtemp("snek5000", __name__))

    def complete_create_default_params(p):
        if hasattr(p.nek.general, "_recorded_user_params"):
            p.nek.general._recorded_user_params.clear()
        p._set_attribs({"prandtl": 0.71, "rayleigh": 1.8e8})
        p._record_nek_user_params({"prandtl": 1, "rayleigh": 2})
        p._set_child("output")
        p.output._set_child("other", {"write_interval": 100})
        p.output.other._record_nek_user_params({"write_interval": 10}, overwrite=True)

        with pytest.raises(ValueError):
            p._change_index_userparams({2: "prandtl"})

        with pytest.raises(ValueError):
            p._change_index_userparams({5: "foo"})

        p._change_index_userparams({2: "prandtl", 1: "rayleigh"})

    from snek5000.solvers.base import Simul

    params = Simul.create_default_params()
    complete_create_default_params(params)

    assert params.nek.general._recorded_user_params == {
        2: "prandtl",
        1: "rayleigh",
        10: "output.other.write_interval",
    }

    params.prandtl = 2
    params.rayleigh = 2e8
    params.output.other.write_interval = 1000

    path_par = tmp_dir / "tmp.par"
    _save_par_file(params, path_par)

    params1 = Simul.create_default_params()
    complete_create_default_params(params1)

    complete_params_from_par_file(params1, path_par)

    assert params1.prandtl == params.prandtl
    assert params1.rayleigh == params.rayleigh
    assert params1.output.other.write_interval == params.output.other.write_interval


def test_str_par_file(sim, tmp_path):
    path = tmp_path / "test.par"
    par_save_par_file = save_par_file_and_read(sim.params, path)
    par_str_par_file = _str_par_file(sim.params)

    assert par_save_par_file == par_str_par_file



import sys
from unittest.mock import patch

import pytest
from conftest import unset_snek_debug
from pymech.neksuite.field import read_header

from snek5000 import load
from snek5000.util.restart import main


def test_restart_command_help():
    with patch.object(sys, "argv", ["snek-restart", "-h"]):
        with pytest.raises(SystemExit):
            main()


@pytest.mark.slow
def test_restart_command_not_enough_args(sim_cbox_executed):
    command = ["snek-restart", sim_cbox_executed.output.path_run, "-oc"]
    with patch.object(sys, "argv", command):
        with pytest.raises(SystemExit):
            main()


@pytest.mark.slow
def test_restart_command_only_check(sim_cbox_executed):
    command = [
        "snek-restart",
        sim_cbox_executed.output.path_run,
        "-oc",
        "--use-checkpoint",
        "1",
        "--new-dir-results",
        "--num-steps",
        "2",
    ]
    with patch.object(sys, "argv", command):
        main()


@pytest.mark.slow
def test_restart_command_only_init(sim_cbox_executed):
    command = [
        "snek-restart",
        sim_cbox_executed.output.path_run,
        "-oi",
        "--use-start-from",
        "-1",
        "--new-dir-results",
        "--end-time",
        "0.015",
    ]
    with patch.object(sys, "argv", command):
        main()


def _test_run(sim_executed, capsys, command, end_time):
    with patch.object(sys, "argv", command):
        with unset_snek_debug():
            main()

    out = capsys.readouterr().out
    path_new = out.split("# To visualize with IPython:\n\ncd ")[-1].split(";")[0]
    sim = load(path_new)
    assert sim_executed.path_run != sim.path_run
    assert (sim.path_run / "session_00").exists()

    header = read_header(sim.output.get_field_file())
    assert header.time == end_time


@pytest.mark.slow
def test_restart_command_run(sim_cbox_executed_readonly, capsys):
    end_time = 0.012
    command = [
        "snek-restart",
        str(sim_cbox_executed_readonly.output.path_run),
        "--use-checkpoint",
        "1",
        "--new-dir-results",
        "--end-time",
        str(end_time),
    ]
    _test_run(sim_cbox_executed_readonly, capsys, command, end_time)


@pytest.mark.slow
def test_restart_command_run_start_from(sim_cbox_executed, capsys):
    end_time = 0.012
    command = [
        "snek-restart",
        sim_cbox_executed.output.path_run,
        "--use-start-from",
        "-1",
        "--new-dir-results",
        "--end-time",
        str(end_time),
    ]
    _test_run(sim_cbox_executed, capsys, command, end_time)



"""Logging
==========
A logger instance (variable :code:`logger`).

"""

import logging
import os

logger = logging.getLogger(__name__)


def create_handler(width=None):
    try:
        from rich.console import Console
        from rich.logging import RichHandler

        console = Console(width=width, stderr=True)
        handler = RichHandler(console=console, show_time=False, show_path=False)
    except ImportError:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    return handler


handler = create_handler()
logger.addHandler(handler)


if os.getenv("SNEK_DEBUG"):
    level = logging.DEBUG
else:
    level = logging.INFO

logger.setLevel(level)



from math import pi

from snek5000.info import InfoSolverMake
from snek5000.params import complete_params_from_par_file
from snek5000.solvers.base import SimulNek


class InfoSolverTGV(InfoSolverMake):
    """Contain the information on a :class:`snek5000_tgv.solver.Simul`
    instance.

    """

    def _init_root(self):
        super()._init_root()
        self.module_name = "snek5000_tgv.solver"
        self.class_name = "Simul"
        self.short_name = "tgv"

        self.classes.Output.module_name = "snek5000_tgv.output"
        self.classes.Output.class_name = "OutputTGV"


class SimulTGV(SimulNek):
    """A solver which compiles and runs using a Snakefile."""

    InfoSolver = InfoSolverTGV

    @classmethod
    def create_default_params(cls):
        """Set default values of parameters as given in reference
        implementation.

        """
        params = super().create_default_params()
        # Re-define default values for parameters here, if necessary

        # Mesh: from ms0.box and SIZE files
        params.oper.Lx = params.oper.Ly = params.oper.Lz = 2 * pi
        params.oper.nx = params.oper.ny = params.oper.nz = 8
        params.oper.elem.order = params.oper.elem.order_out = 8
        params.oper.elem.staggered = False

        # Snek5000 also has a human-friendly interface for userParams## parameter in the par file
        # Here we map the mesh length to userParam01, which is used in subroutine usrdat
        params.oper._record_nek_user_params({"Lx": 1})

        # Read defaults for `params.nek` from `tgv.par.cfg` (original code)
        info_solver = cls.info_solver  # cls.InfoSolver()
        output_cls = info_solver.import_classes()["Output"]
        root = output_cls.get_path_solver_package()
        complete_params_from_par_file(
            params, root / f"{info_solver.short_name}.par.cfg"
        )

        # Boundary conditions
        # Symmetric on top and bottom; periodic on all 6 faces
        params.oper.boundary = ["P"] * 6

        # NOTE: These are only default parameters for a minimal simulation -
        # which can be verify different for another case. Modifications to
        # ``params`` by the user in a separate script are also encouraged.
        return params


Simul = SimulTGV



"""Utility for configuration files"""

import os
import shutil
from pathlib import Path
from socket import gethostname

from snek5000 import get_snek_resource


def ensure_config_file():
    """Ensure that a configuration file is present.

    This function is used by the console script `snek-generate-config`.
    """
    host = os.getenv("SNIC_RESOURCE", os.getenv("GITHUB_WORKFLOW", gethostname()))
    xdg_config = Path(os.path.expandvars(os.getenv("XDG_CONFIG_HOME", "$HOME/.config")))
    configfile_xdg_config_host = xdg_config / f"snek5000/{host}.yml"
    configfile_xdg_config = xdg_config / "snek5000.yml"

    config_files = [configfile_xdg_config_host, configfile_xdg_config]
    config_files = [path for path in config_files if path.exists()]

    if not config_files:
        print("No user config file found")
        configfile_default = Path(get_snek_resource("default_configfile.yml"))
        print(f"Copying {configfile_default} to {configfile_xdg_config}")
        configfile_xdg_config.parent.mkdir(exist_ok=True)
        shutil.copyfile(configfile_default, configfile_xdg_config)
    else:
        print(f"Found configuration file {config_files[0]}")
